# -*- coding: iso-8859-1 -*-
"""
Code for Simplified Good-Turing estimation, particularly for word 
frequencies.  See also Sampson, Empirical Linguistics.

The easiest way to use it is to feed a sequence into 
getGoodTuringEstimateFromSeq and get back the probability
of the unseen species p0 and a dictionary of estimated
probabilites for the sequence elements.

Vgl. Skript: "Der Good-Turing-Schtzer III"
"""

import operator
from math import exp, sqrt
from math import log10 as log

def _getFreqs(seq):
	"""returns a dictionary of frequencies of species in the sequence seq.
	"""
	freqs = {}
	for it in seq:
		freqs[it] = freqs.get(it, 0)+1
	return freqs

def _linreg(data): 
	"""returns the coefficients a,b of the regression line ax+b through
	the list of (x,y) tuples data.
	Adapted from William Park's simple recipies in Python.
	"""
	N = len(data) 
	Sx = Sy = Sxx = Syy = Sxy = 0.0 
	for x, y in data:
		Sx = Sx + x 
		Sy = Sy + y 
		Sxx = Sxx + x*x 
		Syy = Syy + y*y 
		Sxy = Sxy + x*y 
	det = Sxx * N - Sx * Sx 
	a, b = (Sxy * N - Sy * Sx)/det, (Sxx * Sy - Sx * Sxy)/det 
	return a, b

def _smoothData(unsmoothedData):
	"""smoothes the list of tuples unsmoothedData by essentially taking
	the x spacing into account and lowering y values of points the
	neighbours of which are far away.  This is useful for frequencies
	of frequencies, since it blurs out a 1 at a high f.o.f. over
	many frequencies, as it were.  It also takes the logs of both
	smoothed coordinates in order to facilitate linear regression
	for polynomials.
	"""
	xValues = [0]+[f[0] for f in 
		unsmoothedData]+[2*unsmoothedData[-1][0]-unsmoothedData[-2][0]]
	smoothedData = []
	for j in range(len(unsmoothedData)):
		smoothedData.append((log(unsmoothedData[j][0]), 
			log(2*unsmoothedData[j][1]/float((xValues[j+2]-xValues[j])))))
	return smoothedData

def _computeFoFEstimation(fofTable, slope, offset):
	"""returns a dictionary of estimated frequencies for observed frequencies
	for the table of frequency of frequency fofTable and fit parameters
	of Zipf's law slope, offset (use _linreg to estimate these two).
	"""
	rStar = {}
	useSmoothed = 1
	for i in range(len(fofTable)):
		r = fofTable[i][0]
		y = (r+1)*10**(slope*(log((r+1)/float(r))))
		if useSmoothed:
			rStar[r] = y
		else:
			try:
				if fofTable[i][0]+1!=fofTable[i+1][0]:
					useSmoothed = 1
			except IndexError:
				useSmoothed = 1
			if not useSmoothed:
				nr, nrP1 = fofTable[i][1], fofTable[i+1][1]
				x = (r+1)*nrP1/float(nr)
				thresh = 1.96*sqrt((r+1)**2*nrP1/float(nr)**2*(1+nrP1/float(nr)))
				if abs(x-y)>thresh:
					useSmoothed = 1
			if useSmoothed:
				rStar[r] = y
			else:
				rStar[r] = x
	return rStar

def _computeFreqToProb(rToEstFreq, fofTable):
	"""returns p0, the probability of observing unseen species, and
	a mapping from observed frequency to probability for the frequency
	of frequency table fofTable and the mapping from observed frequencies
	to estimated probability rToEstFreq.  The latter can be obtained from
	_computeFoFEstimation.
	"""
	sampleSize = reduce(operator.add, [l[0]*l[1] for l in fofTable])
	p0 = fofTable[0][1]/float(sampleSize)
	reducedSampleSize = float(reduce(operator.add, 
		[rToEstFreq[r]*f for r,f in fofTable]))
	freqToProb = {}
	for r,f in fofTable:
		freqToProb[r] = (1-p0)*rToEstFreq[r]/reducedSampleSize
	return p0, freqToProb

def getGoodTuringEstimate(rawDist):
	"""returns a Good-Turing smoothed version of the dictionary
	rawDist (the keys are the observed frequencies) along with
	the probability of unseen species.
	"""
	fofTable = _getFreqs(rawDist.values()).items()
	fofTable.sort()
	smoothedData = _smoothData(fofTable)
	slope, offset = _linreg(smoothedData)
	freqToEstFreq = _computeFoFEstimation(fofTable, slope, offset)
	p0, freqToProb = _computeFreqToProb(freqToEstFreq, fofTable)
	estDist = {}
	for key in rawDist.keys():
		estDist[key] = freqToProb[rawDist[key]]
	return p0, estDist

def getGoodTuringEstimateFromSeq(seq):
	"""returns a SGT-Estimate of species probabilities in seq together with
	the probability of unseed species.
	"""
	rawDist = _getFreqs(seq)
	return getGoodTuringEstimate(rawDist)


def _test():
	"""This is an example from Sampson.
	"""
	import sys
	fofTable = [(1,120),
		(2,40), (3,24), (4,13), (5,15), (6,5), (7,11), (8,2), (9,2), (10,1),
		(12,3), (14,2), (15,1), (16,1), (17,3), (17,1), (20,3), (21,2), (23,3),
		(24,3), (25,3), (26,2), (27,2), (28,1), (31,2), (32,2), (33,1), (34,2),
		(36,2), (41,3), (43,1), (45,3), (46,1), (47,1), (50,1), (71,1), (84,1),
		(101,1), (105,1), (121,1), (124,1), (146,1), (162,1), (193,1), (199,1),
		(224,1), (226,1), (254,1), (257,1), (339,1), (421,1), (456,1), (481,1),
		(483,1), (1140,1), (1256,1), (1322,1), (1530,1), (2131,1), (2395,1),
		(6925,1), (7846,1),]
	smoothedData = _smoothData(fofTable)
	slope, offset = _linreg(smoothedData)
	freqToEstFreq = _computeFoFEstimation(fofTable, slope, offset)
	p0, est = _computeFreqToProb(freqToEstFreq, fofTable)
	if (abs(p0-0.00388349514563)>1e-10 or 
		abs(est[10]-0.00031152533601934754)>1e-10 or
		abs(est[257]-0.0082975241094069806)>1e-10 or
		abs(est[1256]-0.040600300216825522)>1e-10):
		sys.stderr.write("**** Self test failed! ****\n")


if __name__=="__main__":
	_test()
